use std::path::Path;

use log::debug;
use sqlx::{migrate::MigrateDatabase, sqlite::SqliteRow, Executor, Sqlite, SqlitePool};

use crate::sheet::{EnumSheet, Sheet};

pub struct Database {
    connection: SqlitePool,
}

impl Database {
    pub async fn setup(database_path: impl AsRef<Path>) -> sqlx::Result<Self> {
        let connection =
            Database::create_connection(database_path.as_ref().to_str().unwrap()).await?;
        sqlx::migrate!("./db-migrations").run(&connection).await?;
        Ok(Database { connection })
    }

    async fn create_connection(database_path: &str) -> sqlx::Result<SqlitePool> {
        if !Sqlite::database_exists(database_path)
            .await
            .unwrap_or(false)
        {
            match Sqlite::create_database(database_path).await {
                Ok(_) => debug!("Database created"),
                Err(e) => {
                    debug!("Error creating database: {}", e);
                    return Err(e);
                }
            }
        }

        let connection = SqlitePool::connect(database_path).await?;
        debug!("Connected to database");
        Ok(connection)
    }

    pub async fn insert_sheet(&self, sheet: &EnumSheet) -> sqlx::Result<()> {
        sheet
            .insert_to_database_query()
            .execute(&self.connection)
            .await
            .map(|_| ())
    }

    pub async fn update_sheet_path(&self, sheet: &EnumSheet) -> sqlx::Result<()> {
        sheet
            .update_path_in_database_query()
            .execute(&self.connection)
            .await
            .map(|_| ())
    }

    pub async fn update_sheet_last_opened(&self, sheet: &EnumSheet) -> sqlx::Result<()> {
        sheet
            .update_last_opened_in_database_query()
            .execute(&self.connection)
            .await
            .map(|_| ())
        // TODO: check for success
    }

    pub async fn fetch_all_sheets(&self) -> sqlx::Result<Vec<EnumSheet>> {
        let mut stream = sqlx::query("SELECT * FROM users")
            .map(|row: SqliteRow| {})
            .fetch(&mut conn);
        sqlx::query_as::<_, Sheet>("SELECT * FROM sheets")
            .fetch_all(&self.connection)
            .await
    }

    pub async fn insert_orphan_file(&self, file: &OrphanFile) -> sqlx::Result<i64> {
        sqlx::query(
            "
                INSERT INTO orphan_files (path, file_size, file_hash, last_opened)
                VALUES ($1, $2, $3, $4)
            ",
        )
        .bind(file.path.to_str().unwrap().to_string())
        .bind(file.file_size as i32)
        .bind(file.file_hash.clone())
        .bind(file.last_opened.timestamp())
        .execute(&self.connection)
        .await
        .map(|result| result.last_insert_rowid())
    }

    pub async fn update_orphan_file_path(&self, orphan: &OrphanFile) -> sqlx::Result<()> {
        sqlx::query("UPDATE orphan_files SET path = $1 WHERE id = $2")
            .bind(orphan.path.to_str().unwrap().to_string())
            .bind(orphan.id)
            .execute(&self.connection)
            .await
            .map(|_| ())
    }

    pub async fn update_orphan_last_opened(&self, orphan: &OrphanFile) -> sqlx::Result<()> {
        sqlx::query("UPDATE orphan_files SET last_opened = $1 WHERE id = $2")
            .bind(orphan.last_opened.timestamp())
            .bind(orphan.id)
            .execute(&self.connection)
            .await
            .map(|_| ())
        // TODO: check for success
    }

    pub async fn fetch_all_orphan_files(&self) -> sqlx::Result<Vec<OrphanFile>> {
        sqlx::query_as::<_, OrphanFile>("SELECT * FROM orphan_files")
            .fetch_all(&self.connection)
            .await
    }

    pub fn get_executor(&self) -> E
    where
        E: Executor,
    {
        self.connection
    }
}
